/*
 * rstatd for linux using /proc/stat
 * Copyright (C) 1993 Rudolf Koenig 
 *
 * Adapted for Linux kernel version 0.99.15y and newer by Karl Keyte
 * Date: 94.03.09          e-Mail: KKEYTE@ESOC.BITNET
 *
 * Adapted for Linux kernel version 1.00 and newer by Malcolm Reeves
 * Date: 94.05.24          e-Mail: reeves@pangea.usask.ca
 *
 * Optimizing and network stats (only tested with 1.3.11!) by Michael Rausch
 * Date: 95.07.23          e-Mail: mrausch@ernie.mi.uni-koeln.de
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <rpc/rpc.h>
#include <stdio.h>
#include <fcntl.h>
#include "rstat.h"

struct statstime  s_t;
struct statsswtch s_s;
struct stats      s_o; /* orig */
int havedisk = 0;

static void fill_s_t(void);

struct statstime *
rstatproc_stats_3_svc(void *v, struct svc_req *cl)
{
  fill_s_t();
  return &s_t;
}

struct statsswtch *
rstatproc_stats_2_svc(void *v, struct svc_req *cl)
{
  fill_s_t();
  bcopy((char *)&s_t, (char *)&s_s, sizeof(s_s));
  return &s_s;
}

struct stats *
rstatproc_stats_1_svc(void *v, struct svc_req *cl)
{
  fill_s_t();
  bcopy((char *)&s_t, (char *)&s_o, sizeof(s_o));
  return &s_o;
}

unsigned int *
rstatproc_havedisk_3_svc(void *v, struct svc_req *cl)
{
  fill_s_t();
  return &havedisk;
}

unsigned int *
rstatproc_havedisk_2_svc(void *v, struct svc_req *cl)
{
  return rstatproc_havedisk_3_svc(v, cl);
}

unsigned int *
rstatproc_havedisk_1_svc(void *v, struct svc_req *cl)
{
  return rstatproc_havedisk_3_svc(v, cl);
}

/* FIXME: unsigned/signed overflow (But: It should seldom occur :-) */
static void
fill_s_t(void)
{
  struct timeval tv;
  static int procnetdev_title = 0, procnetdev_line = 0,
	     rp_p = 0, re_p = 0, tp_p = 0, te_p = 0, tc_p = 0;
  int fd, i;
  unsigned long rp,re, tp,te,tc;
  char *buf, *next,
       buffer[PROCNETDEV_LENGTH+8];

  if( (fd = open("/proc/stat", O_RDONLY)) <0)
    return;
  if( (i = read(fd, buffer, sizeof(buffer)-8)) < 0 || i > sizeof(buffer)-8)
  { close(fd); return; }
  memset(&buffer[i], '\0', 8); /* to allow quick jump over "cpu  " etc. */
  close(fd);

  buf = buffer;

#define STRTOL() strtol(buf, &next, 10); if(buf==next) return; buf=next

  buf += strlen("cpu  ");
  s_t.cp_time[0] = STRTOL();
  s_t.cp_time[1] = STRTOL();
  s_t.cp_time[2] = STRTOL();
  s_t.cp_time[3] = STRTOL();

  buf += strlen("\ndisk ");
  s_t.dk_xfer[0] = STRTOL();
  s_t.dk_xfer[1] = STRTOL();
  s_t.dk_xfer[2] = STRTOL();
  s_t.dk_xfer[3] = STRTOL();

  buf += strlen("\npage ");
  s_t.v_pgpgin = STRTOL();
  s_t.v_pgpgout = STRTOL();

  buf += strlen("\nswap ");
  s_t.v_pswpin = STRTOL();
  s_t.v_pswpout = STRTOL();

  buf += strlen("\nintr ");
  s_t.v_intr = STRTOL();

  if( (buf = strchr(buf, '\n')) == NULL ) return; /* skip the interrupts ... */

  buf += strlen("\nctxt ");
  s_t.v_swtch = STRTOL();

  buf += strlen("\nbtime ");
  s_t.boottime.tv_sec = STRTOL();

  gettimeofday(&tv, NULL);
  s_t.curtime.tv_sec = tv.tv_sec;

  /* don't bother with the microseconds */
  s_t.curtime.tv_usec = 0;
  s_t.boottime.tv_usec = 0;

  if( (fd = open("/proc/loadavg", O_RDONLY)) <0)
    return;
  if( (i = read(fd, buffer, sizeof(buffer)-8)) < 0 || i > sizeof(buffer)-8)
  { close(fd); return; }
  buffer[i] = '\0';
  close(fd);

  s_t.avenrun[0] = strtod(buffer, &next) * FSCALE;
  s_t.avenrun[1] = strtod(next, &next) * FSCALE;
  s_t.avenrun[2] = strtod(next, &next) * FSCALE;

  havedisk = 0; /* whatever this is good for ... */

  /*
   *  network stats implemented by Michael Rausch
   *
   *  In the first run, several positions in the output of /proc/net/dev
   *  are calculated and reused later on.
   *  This speeds up the thingy quite a lot, as the while loop is taken 
   *  e.g. 11 times on my machine (loopback + dummy + 4*ppp + 4*slip +
   *  ether). 
   *
   */

  s_t.if_ipackets = s_t.if_ierrors = s_t.if_oerrors = s_t.if_collisions = 
  s_t.if_opackets = 0;

  if( (fd = open("/proc/net/dev", O_RDONLY)) <0)
    return;
  if( (i = read(fd, buffer, sizeof(buffer)-8)) < 0 ||  i > sizeof(buffer)-8)
  { close(fd); return; }
  buffer[i] = '\0';
  close(fd);
 
/*  
 * Skip first two lines, that is ...

Inter-|   Receive                  |  Transmit
 face |packets errs drop fifo frame|packets errs drop fifo colls carrier

 */
  if(procnetdev_title)
    buf = buffer+procnetdev_title;
  else
  {
    if( (buf = strchr(buffer, '\n')) == NULL ) return;
    if( (buf = strchr(buf+1, '\n')) == NULL ) return;
    buf++;
    procnetdev_title = buf-buffer;
  }

#ifndef COUNT_LOOPBACK
  if(procnetdev_line)
    buf += procnetdev_line;
  else
  {
    /* skip first device, that is local loopback (is it always?) */
    if( (buf = strchr(buf+1, '\n')) == NULL ) return;
    buf++;
  }
#endif

  while(*buf) 
  {
    if(procnetdev_line) /* there are cached values, use them */
    {
      rp = strtol(&buf[rp_p], &next, 10);
      if(&buf[rp_p] == next)
      {
        if( (buf = strchr(buf, '\n')) == NULL ) break;
        buf++;
        continue; /* dummy: [etc] */
      }

      re = strtol(&buf[re_p], NULL, 10);
      tp = strtol(&buf[tp_p], NULL, 10);
      te = strtol(&buf[te_p], NULL, 10);
      tc = strtol(&buf[tc_p], NULL, 10);

      buf += procnetdev_line;
    }
    else /* compute the positions and cache them */
    {
      char *startline;

      if( *(buf-1) != '\n')
      {
        if( (buf = strchr(buf, '\n')) == NULL ) break;
        buf++;
      }

      startline = buf;

      if( (buf = strchr(buf, ':')) == NULL ) break;
      buf++;

      rp_p = buf-startline;
      rp = strtol(buf, &next, 10);
      if(buf==next) continue; /* dummy: [etc] */
      buf = next;

      re_p = buf-startline;
      re = strtol(buf, &next, 10);
      if(buf==next) continue;
      buf = next;

      for(i=0; i<3; i++) /* drop, fifo, frame */
      {
        while(isspace(*buf)) buf++;
        while(!isspace(*buf)) buf++;
      }

      tp_p = buf-startline;
      tp = strtol(buf, &next, 10);
      if(buf==next) continue;
      buf = next;

      te_p = buf-startline;
      te = strtol(buf, &next, 10);
      if(buf==next) continue;
      buf = next;

      for(i=0; i<2; i++) /* drop, fifo */
      {
        while(isspace(*buf)) buf++;
        while(!isspace(*buf)) buf++;
      }

      tc_p = buf-startline;
      tc = strtol(buf, &next, 10);
      if(buf==next) continue;
      buf = next;

      /* while(isspace(*buf)) buf++; while(!isspace(*buf)) buf++; / * carrier */

      if( (buf = strchr(buf, '\n')) == NULL ) break;
      buf++;

      procnetdev_line = buf-startline;
    }

    s_t.if_ipackets += rp;
    s_t.if_ierrors += re;
    s_t.if_oerrors += te;
    s_t.if_collisions += tc;
    s_t.if_opackets += tp;
  }

}
